constants = {
    'ONE_TRADING_DAY': {
        'EQUITY': (1/252),
        'CRYPTO': (1/365)
    },
    'ACCURACY': 7,
    'BACKOFF_PERIOD': 30,
    'KEEP_FILE': '.gitkeep',
    'PRICE_YEAR_CUTOFF': 1950,
    'DENOMINATION': 'USD',
    'NPV_DELTA_TOLERANCE': 0.0000001,
    'OPTIMIZATION_METHOD': "SLSQP",
    'SIG_FIGS': 5,
    'LOG_LEVEL': {
        'NONE': 'none',
        'INFO': 'info',
        'DEBUG': 'debug',
        'VERBOSE': 'verbose'
    }
}
